import time
from typing import Dict, List

from cognite.datastudio.exceptions import ModelFailedException


def request(client, path, request_type, data=None):
    project = client._config.project
    url = f"/api/playground/projects/{project}/context/entity_matching{path}"
    if request_type == "GET":
        response = client._api_client._get(url)
    else:
        response = client._api_client._post(url, json=data, headers={"Content-Type": "application/json"})

    return response


class Model:
    def __init__(self, client, model_id):
        self.model_id = model_id
        self.client = client

    def __repr__(self):
        return "Model(id: %d)" % self.model_id

    def predict(self, y: List[str]):
        y = list(y)
        response = request(self.client, f"/{self.model_id}/predict", "POST", data={"items": y})
        data = response.json()
        job_id = data["jobId"]
        while data["status"] != "Completed":
            response = request(self.client, f"/{self.model_id}/predict/{job_id}", "GET")
            data = response.json()
            if data["status"] == "Failed":
                raise ModelFailedException(f"model_id: {self.model_id}")
            time.sleep(1)
        return data["items"]


class EntityMatcher:
    def __init__(self, client):
        self.client = client

    def fit(self, X: List[str]):
        X = list(X)
        response = request(self.client, "/fit", "POST", data={"items": X})

        data = response.json()
        model_id = data["modelId"]
        while data["status"] != "Completed":
            response = request(self.client, f"/{model_id}", "GET")
            data = response.json()
            if data["status"] == "Failed":
                raise ModelFailedException(f"model_id: {model_id}")
            time.sleep(1)
        return Model(self.client, model_id)

    def create_rules(self, matches: List[Dict]):
        matches = list(matches)
        response = request(self.client, "/rules", "POST", data={"items": matches})
        data = response.json()
        job_id = data["jobId"]

        while data["status"] != "Completed":
            response = request(self.client, f"/rules/{job_id}", "GET")
            data = response.json()
            if data["status"] == "Failed":
                raise ModelFailedException(f"job_id: {job_id}")
            time.sleep(1)
        return data["items"]



from typing import List

from cognite.client import CogniteClient
from cognite.datastudio.data_kit import DataKit


class DataPipeline:
    data_kits: List[DataKit] = []

    def load_data_kits(self):
        c = CogniteClient(client_name="datastudio")

        self.data_kits = []

        # Data kit 1
        data_kit = DataKit("Skarv assets and timeseries")
        data_kit.assets = c.assets.list(root_ids=[8129784932439587], limit=5000)
        data_kit.time_series = c.time_series.list(root_asset_ids=[8129784932439587], limit=5000)
        self.data_kits.append(data_kit)

    def create_relations(self, type1, type2, matches):
        print("Done")
        pass


current_data_pipeline = DataPipeline()



from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
import regex as re


def _doRow(row: pd.Series):
    X = row["predicted"]
    Y = row["input"]

    pattern = "[A-Za-z]+|[0-9]+"
    Yalpha = pd.Series(re.findall(pattern, Y))
    Ysepa = pd.Series(re.split(pattern, Y))
    Xalpha = pd.Series(re.findall(pattern, X))
    Xsepa = pd.Series(re.split(pattern, X))
    common, ixX, ixY = np.intersect1d(Xalpha, Yalpha, return_indices=True)
    ixX = pd.Series(ixX)
    ixY = pd.Series(ixY)

    YN = pd.Series(["D" if a.isdigit() else "L" for a in Yalpha])
    YN.loc[ixY] = "[" + YN.loc[ixY] + (1 + ixY.index).astype(str) + "]"
    YNN = [Ysepa[i] + YN[i] for i in range(len(YN))]

    XN = pd.Series(["D" if a.isdigit() else "L" for a in Xalpha])
    XN.loc[ixX] = "[" + XN.loc[ixX] + (1 + ixX.index).astype(str) + "]"
    XNN = [Xsepa[i] + XN[i] for i in range(len(XN))]

    PY = "".join(YNN)
    PX = "".join(XNN)
    return [PY, PX]


def _calcRules(M: pd.DataFrame):
    rule = []
    VAL_Y = M["PY"].value_counts()
    for vc in VAL_Y.index:
        Mred = M[M["PY"] == vc]
        VAL_X = Mred.PX.value_counts()
        for vx in VAL_X.index:
            Mrred = Mred[Mred.PX == vx]
            score = Mrred.score.mean()
            rule.append([vc, vx, VAL_X.loc[vx], score, Mrred.index])

    return pd.DataFrame(
        columns=["Y Pattern", "X Pattern", "num_matches", "avg_score", "match_ix"], data=rule
    ).sort_values(by="num_matches", ascending=False)


def calc_regex_rules(matches: pd.DataFrame):
    A = matches[["predicted", "input"]].apply(_doRow, axis=1, result_type="expand")
    A.columns = ["PY", "PX"]
    A["score"] = matches.score

    return _calcRules(A)


def _get_chunks_and_separators(string: str, pattern: str = "[A-Za-z]+|[0-9]+") -> Tuple[List[str], List[str]]:
    return re.findall(pattern, string), re.split(pattern, string)


def _make_chunks_into_pattern_list(chunks: List[str]) -> List[str]:
    return ["D" if a.isdigit() else "L" for a in chunks]


def _strip_encoded_pattern(string_pattern):
    return re.sub(r"[\[\]\d]", "", string_pattern)


def _get_string_pattern(string: str, pattern: str = "[A-Za-z]+|[0-9]+") -> str:
    """
    Based on the regex pattern for tokens, get the encoded string representing string pattern
    """
    chunks, separators = _get_chunks_and_separators(string, pattern)
    pattern_chunks = _make_chunks_into_pattern_list(chunks)
    return "".join(map(lambda x, y: x + y, separators, pattern_chunks))


def _get_matching_pattern_chunks(string: str, string_pattern: str, pattern: str = "[A-Za-z]+|[0-9]+") -> Optional[str]:
    """
    Based on the pattern for tokens and string_pattern, get the string
    representing tokens values from the matching part of the pattern.
    """
    if _get_string_pattern(string, pattern) != _strip_encoded_pattern(string_pattern):
        return None
    pattern_chunks = list(zip(re.findall(r"[L|D]\d*", string_pattern), re.findall(pattern, string)))
    keep_pattern_chunk = [pattern_chunk for pattern_chunk in pattern_chunks if re.match(r"[L|D]\d+", pattern_chunk[0])]
    keep_pattern_chunk.sort(key=lambda x: x[0])
    return str(keep_pattern_chunk)


def apply_rule(
    input_pattern: str, predict_pattern: str, input_list: List[str], predicted_list: List[str]
) -> pd.DataFrame:
    """
    Matches entries in `input_list` against entries in `predicted_list`. The entries in
    `input_list` and `predicted_list` are chunked according to `input_pattern` and
    `predict_pattern`, respectively, and entries with equal chunks are considered to match.

    Args:
        input_pattern (str): Pattern to which the entries in `input_list` are chunked.
        predict_pattern (str): Pattern to which the entries in `predicted_list` are chunked.
        input_list (List(str)): Entries to match from.
        predicted_list (List(str)): Entries to match to.

    Returns:
        (pandas.DataFrame): Data frame with one row per match.
    """

    df_input = pd.DataFrame(input_list)
    df_input.columns = ["input"]
    df_predicted = pd.DataFrame(predicted_list)
    df_predicted.columns = ["predicted"]

    matching_chunks_column = "matching_chunks"
    df_input[matching_chunks_column] = df_input["input"].apply(
        _get_matching_pattern_chunks, string_pattern=input_pattern
    )
    df_predicted[matching_chunks_column] = df_predicted["predicted"].apply(
        _get_matching_pattern_chunks, string_pattern=predict_pattern
    )

    df_matches = df_input.merge(df_predicted, on=matching_chunks_column)
    df_matches = df_matches.dropna()
    return df_matches[["input", "predicted", matching_chunks_column]]



import json
from typing import List, Optional

from cognite.client.data_classes.assets import AssetList
from cognite.client.data_classes.files import FileMetadataList
from cognite.client.data_classes.time_series import TimeSeriesList


class DataKit:
    assets: Optional[List[AssetList]] = None
    time_series: Optional[List[TimeSeriesList]] = None
    files: Optional[List[FileMetadataList]] = None
    name: str = "Data kit"

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        obj = {"name": self.name}
        if self.assets:
            obj["assets"] = "%d assets" % len(self.assets)
        if self.time_series:
            obj["time_series"] = "%d time series" % len(self.time_series)
        if self.files:
            obj["files"] = "%d files" % len(self.files)
        return "<DataKit>: " + json.dumps(obj)
